@@ -1,5 +1,5 @@
 # orm/interfaces.py
-# Copyright (C) 2005-2016 the SQLAlchemy authors and contributors
+# Copyright (C) 2005-2019 the SQLAlchemy authors and contributors
 # <see AUTHORS file>
 #
 # This module is part of SQLAlchemy and is released under
@@ -20,33 +20,40 @@ are exposed when inspecting mappings.
 
 from __future__ import absolute_import
 
+import collections
+
+from . import path_registry
+from .base import _MappedAttribute  # noqa
+from .base import EXT_CONTINUE
+from .base import EXT_STOP
+from .base import InspectionAttr  # noqa
+from .base import InspectionAttrInfo  # noqa
+from .base import MANYTOMANY
+from .base import MANYTOONE
+from .base import NOT_EXTENSION
+from .base import ONETOMANY
+from .. import inspect
 from .. import util
 from ..sql import operators
-from .base import (ONETOMANY, MANYTOONE, MANYTOMANY,
-                   EXT_CONTINUE, EXT_STOP, NOT_EXTENSION)
-from .base import (InspectionAttr, InspectionAttr,
-    InspectionAttrInfo, _MappedAttribute)
-import collections
-from .. import inspect
 
 # imported later
 MapperExtension = SessionExtension = AttributeExtension = None
 
 __all__ = (
-    'AttributeExtension',
-    'EXT_CONTINUE',
-    'EXT_STOP',
-    'ONETOMANY',
-    'MANYTOMANY',
-    'MANYTOONE',
-    'NOT_EXTENSION',
-    'LoaderStrategy',
-    'MapperExtension',
-    'MapperOption',
-    'MapperProperty',
-    'PropComparator',
-    'SessionExtension',
-    'StrategizedProperty',
+    "AttributeExtension",
+    "EXT_CONTINUE",
+    "EXT_STOP",
+    "ONETOMANY",
+    "MANYTOMANY",
+    "MANYTOONE",
+    "NOT_EXTENSION",
+    "LoaderStrategy",
+    "MapperExtension",
+    "MapperOption",
+    "MapperProperty",
+    "PropComparator",
+    "SessionExtension",
+    "StrategizedProperty",
 )
 
 
@@ -63,8 +70,11 @@ class MapperProperty(_MappedAttribute, InspectionAttr, util.MemoizedSlots):
     """
 
     __slots__ = (
-        '_configure_started', '_configure_finished', 'parent', 'key',
-        'info'
+        "_configure_started",
+        "_configure_finished",
+        "parent",
+        "key",
+        "info",
     )
 
     cascade = frozenset()
@@ -117,15 +127,17 @@ class MapperProperty(_MappedAttribute, InspectionAttr, util.MemoizedSlots):
 
         """
 
-    def create_row_processor(self, context, path,
-                             mapper, result, adapter, populators):
+    def create_row_processor(
+        self, context, path, mapper, result, adapter, populators
+    ):
         """Produce row processing functions and append to the given
         set of populators lists.
 
         """
 
-    def cascade_iterator(self, type_, state, visited_instances=None,
-                         halt_on=None):
+    def cascade_iterator(
+        self, type_, state, visited_instances=None, halt_on=None
+    ):
         """Iterate through instances related to the given instance for
         a particular 'cascade', starting with this MapperProperty.
 
@@ -233,21 +245,32 @@ class MapperProperty(_MappedAttribute, InspectionAttr, util.MemoizedSlots):
 
         """
 
-    def merge(self, session, source_state, source_dict, dest_state,
-              dest_dict, load, _recursive, _resolve_conflict_map):
+    def merge(
+        self,
+        session,
+        source_state,
+        source_dict,
+        dest_state,
+        dest_dict,
+        load,
+        _recursive,
+        _resolve_conflict_map,
+    ):
         """Merge the attribute represented by this ``MapperProperty``
         from source to destination object.
 
         """
 
     def __repr__(self):
-        return '<%s at 0x%x; %s>' % (
+        return "<%s at 0x%x; %s>" % (
             self.__class__.__name__,
-            id(self), getattr(self, 'key', 'no key'))
+            id(self),
+            getattr(self, "key", "no key"),
+        )
 
 
 class PropComparator(operators.ColumnOperators):
-    """Defines SQL operators for :class:`.MapperProperty` objects.
+    r"""Defines SQL operators for :class:`.MapperProperty` objects.
 
     SQLAlchemy allows for operators to
     be redefined at both the Core and ORM level.  :class:`.PropComparator`
@@ -273,9 +296,9 @@ class PropComparator(operators.ColumnOperators):
 
         # definition of custom PropComparator subclasses
 
-        from sqlalchemy.orm.properties import \\
-                                ColumnProperty,\\
-                                CompositeProperty,\\
+        from sqlalchemy.orm.properties import \
+                                ColumnProperty,\
+                                CompositeProperty,\
                                 RelationshipProperty
 
         class MyColumnComparator(ColumnProperty.Comparator):
@@ -334,7 +357,7 @@ class PropComparator(operators.ColumnOperators):
 
     """
 
-    __slots__ = 'prop', 'property', '_parententity', '_adapt_to_entity'
+    __slots__ = "prop", "property", "_parententity", "_adapt_to_entity"
 
     def __init__(self, prop, parentmapper, adapt_to_entity=None):
         self.prop = self.property = prop
@@ -347,6 +370,9 @@ class PropComparator(operators.ColumnOperators):
     def _query_clause_element(self):
         return self.__clause_element__()
 
+    def _bulk_update_tuples(self, value):
+        return [(self.__clause_element__(), value)]
+
     def adapt_to_entity(self, adapt_to_entity):
         """Return a copy of this PropComparator which will use the given
         :class:`.AliasedInsp` to produce corresponding expressions.
@@ -387,26 +413,29 @@ class PropComparator(operators.ColumnOperators):
         return a.of_type(class_)
 
     def of_type(self, class_):
-        """Redefine this object in terms of a polymorphic subclass.
+        r"""Redefine this object in terms of a polymorphic subclass.
 
         Returns a new PropComparator from which further criterion can be
         evaluated.
 
         e.g.::
 
-            query.join(Company.employees.of_type(Engineer)).\\
+            query.join(Company.employees.of_type(Engineer)).\
                filter(Engineer.name=='foo')
 
         :param \class_: a class or mapper indicating that criterion will be
             against this specific subclass.
 
+        .. seealso::
+
+            :ref:`inheritance_of_type`
 
         """
 
         return self.operate(PropComparator.of_type_op, class_)
 
     def any(self, criterion=None, **kwargs):
-        """Return true if this collection contains any member that meets the
+        r"""Return true if this collection contains any member that meets the
         given criterion.
 
         The usual implementation of ``any()`` is
@@ -424,7 +453,7 @@ class PropComparator(operators.ColumnOperators):
         return self.operate(PropComparator.any_op, criterion, **kwargs)
 
     def has(self, criterion=None, **kwargs):
-        """Return true if this element references a member which meets the
+        r"""Return true if this element references a member which meets the
         given criterion.
 
         The usual implementation of ``has()`` is
@@ -459,10 +488,30 @@ class StrategizedProperty(MapperProperty):
 
     """
 
-    __slots__ = '_strategies', 'strategy'
+    __slots__ = (
+        "_strategies",
+        "strategy",
+        "_wildcard_token",
+        "_default_path_loader_key",
+    )
 
     strategy_wildcard_key = None
 
+    def _memoized_attr__wildcard_token(self):
+        return (
+            "%s:%s"
+            % (self.strategy_wildcard_key, path_registry._WILDCARD_TOKEN),
+        )
+
+    def _memoized_attr__default_path_loader_key(self):
+        return (
+            "loader",
+            (
+                "%s:%s"
+                % (self.strategy_wildcard_key, path_registry._DEFAULT_TOKEN),
+            ),
+        )
+
     def _get_context_loader(self, context, path):
         load = None
 
@@ -475,7 +524,7 @@ class StrategizedProperty(MapperProperty):
         for path_key in (
             search_path._loader_key,
             search_path._wildcard_path_loader_key,
-            search_path._default_path_loader_key
+            search_path._default_path_loader_key,
         ):
             if path_key in context.attributes:
                 load = context.attributes[path_key]
@@ -488,12 +537,12 @@ class StrategizedProperty(MapperProperty):
             return self._strategies[key]
         except KeyError:
             cls = self._strategy_lookup(*key)
-            self._strategies[key] = self._strategies[
-                cls] = strategy = cls(self, key)
+            self._strategies[key] = self._strategies[cls] = strategy = cls(
+                self, key
+            )
             return strategy
 
-    def setup(
-            self, context, entity, path, adapter, **kwargs):
+    def setup(self, context, entity, path, adapter, **kwargs):
         loader = self._get_context_loader(context, path)
         if loader and loader.strategy:
             strat = self._get_strategy(loader.strategy)
@@ -502,24 +551,26 @@ class StrategizedProperty(MapperProperty):
         strat.setup_query(context, entity, path, loader, adapter, **kwargs)
 
     def create_row_processor(
-            self, context, path, mapper,
-            result, adapter, populators):
+        self, context, path, mapper, result, adapter, populators
+    ):
         loader = self._get_context_loader(context, path)
         if loader and loader.strategy:
             strat = self._get_strategy(loader.strategy)
         else:
             strat = self.strategy
         strat.create_row_processor(
-            context, path, loader,
-            mapper, result, adapter, populators)
+            context, path, loader, mapper, result, adapter, populators
+        )
 
     def do_init(self):
         self._strategies = {}
         self.strategy = self._get_strategy(self.strategy_key)
 
     def post_instrument_class(self, mapper):
-        if not self.parent.non_primary and \
-                not mapper.class_manager._attr_has_impl(self.key):
+        if (
+            not self.parent.non_primary
+            and not mapper.class_manager._attr_has_impl(self.key)
+        ):
             self.strategy.init_class_attribute(mapper)
 
     _all_strategies = collections.defaultdict(dict)
@@ -529,12 +580,13 @@ class StrategizedProperty(MapperProperty):
         def decorate(dec_cls):
             # ensure each subclass of the strategy has its
             # own _strategy_keys collection
-            if '_strategy_keys' not in dec_cls.__dict__:
+            if "_strategy_keys" not in dec_cls.__dict__:
                 dec_cls._strategy_keys = []
             key = tuple(sorted(kw.items()))
             cls._all_strategies[cls][key] = dec_cls
             dec_cls._strategy_keys.append(key)
             return dec_cls
+
         return decorate
 
     @classmethod
@@ -575,6 +627,56 @@ class MapperOption(object):
 
         self.process_query(query)
 
+    def _generate_cache_key(self, path):
+        """Used by the "baked lazy loader" to see if this option can be cached.
+
+        The "baked lazy loader" refers to the :class:`.Query` that is
+        produced during a lazy load operation for a mapped relationship.
+        It does not yet apply to the "lazy" load operation for deferred
+        or expired column attributes, however this may change in the future.
+
+        This loader generates SQL for a query only once and attempts to  cache
+        it; from that point on, if the SQL has been cached it will no longer
+        run the :meth:`.Query.options` method of the :class:`.Query`.   The
+        :class:`.MapperOption` object that wishes to participate within a lazy
+        load operation therefore needs to tell the baked loader that it either
+        needs to forego this caching, or that it needs to include the state of
+        the :class:`.MapperOption` itself as part of its cache key, otherwise
+        SQL or other query state that has been affected by the
+        :class:`.MapperOption` may be cached in place of a query that does not
+        include these modifications, or the option may not be invoked at all.
+
+        By default, this method returns the value ``False``, which means
+        the :class:`.BakedQuery` generated by the lazy loader will
+        not cache the SQL when this :class:`.MapperOption` is present.
+        This is the safest option and ensures both that the option is
+        invoked every time, and also that the cache isn't filled up with
+        an unlimited number of :class:`.Query` objects for an unlimited
+        number of :class:`.MapperOption` objects.
+
+        .. versionchanged:: 1.2.8 the default return value of
+           :meth:`.MapperOption._generate_cache_key` is False; previously it
+           was ``None`` indicating "safe to cache, don't include as part of
+           the cache key"
+
+        To enable caching of :class:`.Query` objects within lazy loaders, a
+        given :class:`.MapperOption` that returns a cache key must return a key
+        that uniquely identifies the complete state of this option, which will
+        match any other :class:`.MapperOption` that itself retains the
+        identical state.  This includes path options, flags, etc.    It should
+        be a state that is repeatable and part of a limited set of possible
+        options.
+
+        If the :class:`.MapperOption` does not apply to the given path and
+        would not affect query results on such a path, it should return None,
+        indicating the :class:`.Query` is safe to cache for this given
+        loader path and that this :class:`.MapperOption` need not be
+        part of the cache key.
+
+
+        """
+        return False
+
 
 class LoaderStrategy(object):
     """Describe the loading behavior of a StrategizedProperty object.
@@ -600,8 +702,14 @@ class LoaderStrategy(object):
 
     """
 
-    __slots__ = 'parent_property', 'is_class_level', 'parent', 'key', \
-        'strategy_key', 'strategy_opts'
+    __slots__ = (
+        "parent_property",
+        "is_class_level",
+        "parent",
+        "key",
+        "strategy_key",
+        "strategy_opts",
+    )
 
     def __init__(self, parent, strategy_key):
         self.parent_property = parent
@@ -624,8 +732,9 @@ class LoaderStrategy(object):
 
         """
 
-    def create_row_processor(self, context, path, loadopt, mapper,
-                             result, adapter, populators):
+    def create_row_processor(
+        self, context, path, loadopt, mapper, result, adapter, populators
+    ):
         """Establish row processing functions for a given QueryContext.
 
         This method fulfills the contract specified by
