@@ -1,5 +1,5 @@
 # sql/default_comparator.py
-# Copyright (C) 2005-2017 the SQLAlchemy authors and contributors
+# Copyright (C) 2005-2019 the SQLAlchemy authors and contributors
 # <see AUTHORS file>
 #
 # This module is part of SQLAlchemy and is released under
@@ -8,21 +8,45 @@
 """Default implementation of SQL comparison operations.
 """
 
-from .. import exc, util
-from . import type_api
 from . import operators
-from .elements import BindParameter, True_, False_, BinaryExpression, \
-    Null, _const_expr, _clause_element_as_expr, \
-    ClauseList, ColumnElement, TextClause, UnaryExpression, \
-    collate, _is_literal, _literal_as_text, ClauseElement, and_, or_, \
-    Slice, Visitable, _literal_as_binds
-from .selectable import SelectBase, Alias, Selectable, ScalarSelect
+from . import type_api
+from .elements import _clause_element_as_expr
+from .elements import _const_expr
+from .elements import _is_literal
+from .elements import _literal_as_text
+from .elements import and_
+from .elements import BinaryExpression
+from .elements import BindParameter
+from .elements import ClauseElement
+from .elements import ClauseList
+from .elements import collate
+from .elements import CollectionAggregate
+from .elements import ColumnElement
+from .elements import False_
+from .elements import Null
+from .elements import or_
+from .elements import TextClause
+from .elements import True_
+from .elements import UnaryExpression
+from .elements import Visitable
+from .selectable import Alias
+from .selectable import ScalarSelect
+from .selectable import Selectable
+from .selectable import SelectBase
+from .. import exc
+from .. import util
 
 
-def _boolean_compare(expr, op, obj, negate=None, reverse=False,
-                     _python_is_types=(util.NoneType, bool),
-                     result_type = None,
-                     **kwargs):
+def _boolean_compare(
+    expr,
+    op,
+    obj,
+    negate=None,
+    reverse=False,
+    _python_is_types=(util.NoneType, bool),
+    result_type=None,
+    **kwargs
+):
 
     if result_type is None:
         result_type = type_api.BOOLEANTYPE
@@ -32,53 +56,76 @@ def _boolean_compare(expr, op, obj, negate=None, reverse=False,
         # allow x ==/!= True/False to be treated as a literal.
         # this comes out to "== / != true/false" or "1/0" if those
         # constants aren't supported and works on all platforms
-        if op in (operators.eq, operators.ne) and \
-                isinstance(obj, (bool, True_, False_)):
-            return BinaryExpression(expr,
-                                    _literal_as_text(obj),
-                                    op,
-                                    type_=result_type,
-                                    negate=negate, modifiers=kwargs)
+        if op in (operators.eq, operators.ne) and isinstance(
+            obj, (bool, True_, False_)
+        ):
+            return BinaryExpression(
+                expr,
+                _literal_as_text(obj),
+                op,
+                type_=result_type,
+                negate=negate,
+                modifiers=kwargs,
+            )
         elif op in (operators.is_distinct_from, operators.isnot_distinct_from):
-            return BinaryExpression(expr,
-                                    _literal_as_text(obj),
-                                    op,
-                                    type_=result_type,
-                                    negate=negate, modifiers=kwargs)
+            return BinaryExpression(
+                expr,
+                _literal_as_text(obj),
+                op,
+                type_=result_type,
+                negate=negate,
+                modifiers=kwargs,
+            )
         else:
             # all other None/True/False uses IS, IS NOT
             if op in (operators.eq, operators.is_):
-                return BinaryExpression(expr, _const_expr(obj),
-                                        operators.is_,
-                                        negate=operators.isnot)
+                return BinaryExpression(
+                    expr,
+                    _const_expr(obj),
+                    operators.is_,
+                    negate=operators.isnot,
+                    type_=result_type,
+                )
             elif op in (operators.ne, operators.isnot):
-                return BinaryExpression(expr, _const_expr(obj),
-                                        operators.isnot,
-                                        negate=operators.is_)
+                return BinaryExpression(
+                    expr,
+                    _const_expr(obj),
+                    operators.isnot,
+                    negate=operators.is_,
+                    type_=result_type,
+                )
             else:
                 raise exc.ArgumentError(
                     "Only '=', '!=', 'is_()', 'isnot()', "
                     "'is_distinct_from()', 'isnot_distinct_from()' "
-                    "operators can be used with None/True/False")
+                    "operators can be used with None/True/False"
+                )
     else:
         obj = _check_literal(expr, op, obj)
 
     if reverse:
-        return BinaryExpression(obj,
-                                expr,
-                                op,
-                                type_=result_type,
-                                negate=negate, modifiers=kwargs)
+        return BinaryExpression(
+            obj, expr, op, type_=result_type, negate=negate, modifiers=kwargs
+        )
     else:
-        return BinaryExpression(expr,
-                                obj,
-                                op,
-                                type_=result_type,
-                                negate=negate, modifiers=kwargs)
+        return BinaryExpression(
+            expr, obj, op, type_=result_type, negate=negate, modifiers=kwargs
+        )
 
 
-def _binary_operate(expr, op, obj, reverse=False, result_type=None,
-                    **kw):
+def _custom_op_operate(expr, op, obj, reverse=False, result_type=None, **kw):
+    if result_type is None:
+        if op.return_type:
+            result_type = op.return_type
+        elif op.is_comparison:
+            result_type = type_api.BOOLEANTYPE
+
+    return _binary_operate(
+        expr, op, obj, reverse=reverse, result_type=result_type, **kw
+    )
+
+
+def _binary_operate(expr, op, obj, reverse=False, result_type=None, **kw):
     obj = _check_literal(expr, op, obj)
 
     if reverse:
@@ -88,10 +135,10 @@ def _binary_operate(expr, op, obj, reverse=False, result_type=None,
 
     if result_type is None:
         op, result_type = left.comparator._adapt_expression(
-            op, right.comparator)
+            op, right.comparator
+        )
 
-    return BinaryExpression(
-        left, right, op, type_=result_type, modifiers=kw)
+    return BinaryExpression(left, right, op, type_=result_type, modifiers=kw)
 
 
 def _conjunction_operate(expr, op, other, **kw):
@@ -111,8 +158,7 @@ def _in_impl(expr, op, seq_or_selectable, negate_op, **kw):
     seq_or_selectable = _clause_element_as_expr(seq_or_selectable)
 
     if isinstance(seq_or_selectable, ScalarSelect):
-        return _boolean_compare(expr, op, seq_or_selectable,
-                                negate=negate_op)
+        return _boolean_compare(expr, op, seq_or_selectable, negate=negate_op)
     elif isinstance(seq_or_selectable, SelectBase):
 
         # TODO: if we ever want to support (x, y, z) IN (select x,
@@ -121,16 +167,27 @@ def _in_impl(expr, op, seq_or_selectable, negate_op, **kw):
         # does not export itself as a FROM clause
 
         return _boolean_compare(
-            expr, op, seq_or_selectable.as_scalar(),
-            negate=negate_op, **kw)
+            expr, op, seq_or_selectable.as_scalar(), negate=negate_op, **kw
+        )
     elif isinstance(seq_or_selectable, (Selectable, TextClause)):
-        return _boolean_compare(expr, op, seq_or_selectable,
-                                negate=negate_op, **kw)
+        return _boolean_compare(
+            expr, op, seq_or_selectable, negate=negate_op, **kw
+        )
     elif isinstance(seq_or_selectable, ClauseElement):
-        raise exc.InvalidRequestError(
-            'in_() accepts'
-            ' either a list of expressions '
-            'or a selectable: %r' % seq_or_selectable)
+        if (
+            isinstance(seq_or_selectable, BindParameter)
+            and seq_or_selectable.expanding
+        ):
+            return _boolean_compare(
+                expr, op, seq_or_selectable, negate=negate_op
+            )
+        else:
+            raise exc.InvalidRequestError(
+                "in_() accepts"
+                " either a list of expressions, "
+                'a selectable, or an "expanding" bound parameter: %r'
+                % seq_or_selectable
+            )
 
     # Handle non selectable arguments as sequences
     args = []
@@ -138,35 +195,26 @@ def _in_impl(expr, op, seq_or_selectable, negate_op, **kw):
         if not _is_literal(o):
             if not isinstance(o, operators.ColumnOperators):
                 raise exc.InvalidRequestError(
-                    'in_() accepts'
-                    ' either a list of expressions '
-                    'or a selectable: %r' % o)
+                    "in_() accepts"
+                    " either a list of expressions, "
+                    'a selectable, or an "expanding" bound parameter: %r' % o
+                )
         elif o is None:
             o = Null()
         else:
             o = expr._bind_param(op, o)
         args.append(o)
+
     if len(args) == 0:
+        op, negate_op = (
+            (operators.empty_in_op, operators.empty_notin_op)
+            if op is operators.in_op
+            else (operators.empty_notin_op, operators.empty_in_op)
+        )
 
-        # Special case handling for empty IN's, behave like
-        # comparison against zero row selectable.  We use != to
-        # build the contradiction as it handles NULL values
-        # appropriately, i.e. "not (x IN ())" should not return NULL
-        # values for x.
-
-        util.warn('The IN-predicate on "%s" was invoked with an '
-                  'empty sequence. This results in a '
-                  'contradiction, which nonetheless can be '
-                  'expensive to evaluate.  Consider alternative '
-                  'strategies for improved performance.' % expr)
-        if op is operators.in_op:
-            return expr != expr
-        else:
-            return expr == expr
-
-    return _boolean_compare(expr, op,
-                            ClauseList(*args).self_group(against=op),
-                            negate=negate_op)
+    return _boolean_compare(
+        expr, op, ClauseList(*args).self_group(against=op), negate=negate_op
+    )
 
 
 def _getitem_impl(expr, op, other, **kw):
@@ -178,13 +226,14 @@ def _getitem_impl(expr, op, other, **kw):
 
 
 def _unsupported_impl(expr, op, *arg, **kw):
-    raise NotImplementedError("Operator '%s' is not supported on "
-                              "this expression" % op.__name__)
+    raise NotImplementedError(
+        "Operator '%s' is not supported on " "this expression" % op.__name__
+    )
 
 
 def _inv_impl(expr, op, **kw):
     """See :meth:`.ColumnOperators.__inv__`."""
-    if hasattr(expr, 'negation_clause'):
+    if hasattr(expr, "negation_clause"):
         return expr.negation_clause
     else:
         return expr._negate()
@@ -199,20 +248,22 @@ def _match_impl(expr, op, other, **kw):
     """See :meth:`.ColumnOperators.match`."""
 
     return _boolean_compare(
-        expr, operators.match_op,
-        _check_literal(
-            expr, operators.match_op, other),
+        expr,
+        operators.match_op,
+        _check_literal(expr, operators.match_op, other),
         result_type=type_api.MATCHTYPE,
         negate=operators.notmatch_op
-        if op is operators.match_op else operators.match_op,
+        if op is operators.match_op
+        else operators.match_op,
         **kw
     )
 
 
 def _distinct_impl(expr, op, **kw):
     """See :meth:`.ColumnOperators.distinct`."""
-    return UnaryExpression(expr, operator=operators.distinct_op,
-                           type_=expr.type)
+    return UnaryExpression(
+        expr, operator=operators.distinct_op, type_=expr.type
+    )
 
 
 def _between_impl(expr, op, cleft, cright, **kw):
@@ -223,17 +274,21 @@ def _between_impl(expr, op, cleft, cright, **kw):
             _check_literal(expr, operators.and_, cleft),
             _check_literal(expr, operators.and_, cright),
             operator=operators.and_,
-            group=False, group_contents=False),
+            group=False,
+            group_contents=False,
+        ),
         op,
         negate=operators.notbetween_op
         if op is operators.between_op
         else operators.between_op,
-        modifiers=kw)
+        modifiers=kw,
+    )
 
 
 def _collate_impl(expr, op, other, **kw):
     return collate(expr, other)
 
+
 # a mapping of operators with the method they use, along with
 # their negated operator for comparison operators
 operator_lookup = {
@@ -246,10 +301,12 @@ operator_lookup = {
     "div": (_binary_operate,),
     "mod": (_binary_operate,),
     "truediv": (_binary_operate,),
-    "custom_op": (_binary_operate,),
-    "json_path_getitem_op": (_binary_operate, ),
-    "json_getitem_op": (_binary_operate, ),
+    "custom_op": (_custom_op_operate,),
+    "json_path_getitem_op": (_binary_operate,),
+    "json_getitem_op": (_binary_operate,),
     "concat_op": (_binary_operate,),
+    "any_op": (_scalar, CollectionAggregate._create_any),
+    "all_op": (_scalar, CollectionAggregate._create_all),
     "lt": (_boolean_compare, operators.ge),
     "le": (_boolean_compare, operators.gt),
     "ne": (_boolean_compare, operators.eq),
@@ -277,8 +334,8 @@ operator_lookup = {
     "match_op": (_match_impl,),
     "notmatch_op": (_match_impl,),
     "distinct_op": (_distinct_impl,),
-    "between_op": (_between_impl, ),
-    "notbetween_op": (_between_impl, ),
+    "between_op": (_between_impl,),
+    "notbetween_op": (_between_impl,),
     "neg": (_neg_impl,),
     "getitem": (_getitem_impl,),
     "lshift": (_unsupported_impl,),
@@ -289,12 +346,11 @@ operator_lookup = {
 
 def _check_literal(expr, operator, other, bindparam_type=None):
     if isinstance(other, (ColumnElement, TextClause)):
-        if isinstance(other, BindParameter) and \
-                other.type._isnull:
+        if isinstance(other, BindParameter) and other.type._isnull:
             other = other._clone()
             other.type = expr.type
         return other
-    elif hasattr(other, '__clause_element__'):
+    elif hasattr(other, "__clause_element__"):
         other = other.__clause_element__()
     elif isinstance(other, type_api.TypeEngine.Comparator):
         other = other.expr
@@ -305,4 +361,3 @@ def _check_literal(expr, operator, other, bindparam_type=None):
         return expr._bind_param(operator, other, type_=bindparam_type)
     else:
         return other
-
