@@ -1,5 +1,5 @@
 # sql/operators.py
-# Copyright (C) 2005-2018 the SQLAlchemy authors and contributors
+# Copyright (C) 2005-2019 the SQLAlchemy authors and contributors
 # <see AUTHORS file>
 #
 # This module is part of SQLAlchemy and is released under
@@ -10,12 +10,28 @@
 
 """Defines operators used in SQL expressions."""
 
+from operator import add
+from operator import and_
+from operator import contains
+from operator import eq
+from operator import ge
+from operator import getitem
+from operator import gt
+from operator import inv
+from operator import le
+from operator import lshift
+from operator import lt
+from operator import mod
+from operator import mul
+from operator import ne
+from operator import neg
+from operator import or_
+from operator import rshift
+from operator import sub
+from operator import truediv
+
 from .. import util
 
-from operator import (
-    and_, or_, inv, add, mul, sub, mod, truediv, lt, le, ne, gt, ge, eq, neg,
-    getitem, lshift, rshift, contains
-)
 
 if util.py2k:
     from operator import div
@@ -37,6 +53,7 @@ class Operators(object):
     :class:`.ColumnOperators`.
 
     """
+
     __slots__ = ()
 
     def __and__(self, other):
@@ -105,8 +122,8 @@ class Operators(object):
         return self.operate(inv)
 
     def op(
-            self, opstring, precedence=0, is_comparison=False,
-            return_type=None):
+        self, opstring, precedence=0, is_comparison=False, return_type=None
+    ):
         """produce a generic operator function.
 
         e.g.::
@@ -168,6 +185,7 @@ class Operators(object):
 
         def against(other):
             return operator(self, other)
+
         return against
 
     def bool_op(self, opstring, precedence=0):
@@ -247,12 +265,18 @@ class custom_op(object):
         :meth:`.Operators.bool_op`
 
     """
-    __name__ = 'custom_op'
+
+    __name__ = "custom_op"
 
     def __init__(
-            self, opstring, precedence=0, is_comparison=False,
-            return_type=None, natural_self_precedent=False,
-            eager_grouping=False):
+        self,
+        opstring,
+        precedence=0,
+        is_comparison=False,
+        return_type=None,
+        natural_self_precedent=False,
+        eager_grouping=False,
+    ):
         self.opstring = opstring
         self.precedence = precedence
         self.is_comparison = is_comparison
@@ -263,8 +287,7 @@ class custom_op(object):
         )
 
     def __eq__(self, other):
-        return isinstance(other, custom_op) and \
-            other.opstring == self.opstring
+        return isinstance(other, custom_op) and other.opstring == self.opstring
 
     def __hash__(self):
         return id(self)
@@ -494,20 +517,79 @@ class ColumnOperators(Operators):
     def in_(self, other):
         """Implement the ``in`` operator.
 
-        In a column context, produces the clause ``a IN other``.
-        "other" may be a tuple/list of column expressions,
-        or a :func:`~.expression.select` construct.
+        In a column context, produces the clause ``column IN <other>``.
 
-        In the case that ``other`` is an empty sequence, the compiler
-        produces an "empty in" expression.   This defaults to the
-        expression "1 != 1" to produce false in all cases.  The
-        :paramref:`.create_engine.empty_in_strategy` may be used to
-        alter this behavior.
+        The given parameter ``other`` may be:
 
-        .. versionchanged:: 1.2  The :meth:`.ColumnOperators.in_` and
-           :meth:`.ColumnOperators.notin_` operators
-           now produce a "static" expression for an empty IN sequence
-           by default.
+        * A list of literal values, e.g.::
+
+            stmt.where(column.in_([1, 2, 3]))
+
+          In this calling form, the list of items is converted to a set of
+          bound parameters the same length as the list given::
+
+            WHERE COL IN (?, ?, ?)
+
+        * An empty list, e.g.::
+
+            stmt.where(column.in_([]))
+
+          In this calling form, the expression renders a "false" expression,
+          e.g.::
+
+            WHERE 1 != 1
+
+          This "false" expression has historically had different behaviors
+          in older SQLAlchemy versions, see
+          :paramref:`.create_engine.empty_in_strategy` for behavioral options.
+
+          .. versionchanged:: 1.2 simplified the behavior of "empty in"
+             expressions
+
+        * A bound parameter, e.g. :func:`.bindparam`, may be used if it
+          includes the :paramref:`.bindparam.expanding` flag::
+
+            stmt.where(column.in_(bindparam('value', expanding=True)))
+
+          In this calling form, the expression renders a special non-SQL
+          placeholder expression that looks like::
+
+            WHERE COL IN ([EXPANDING_value])
+
+          This placeholder expression is intercepted at statement execution
+          time to be converted into the variable number of bound parameter
+          form illustrated earlier.   If the statement were executed as::
+
+            connection.execute(stmt, {"value": [1, 2, 3]})
+
+          The database would be passed a bound parameter for each value::
+
+            WHERE COL IN (?, ?, ?)
+
+          .. versionadded:: 1.2 added "expanding" bound parameters
+
+          The "expanding" feature in version 1.2 of SQLAlchemy does not
+          support passing an empty list as a parameter value; however,
+          version 1.3 does support this.
+
+        * a :func:`.select` construct, which is usually a correlated
+          scalar select::
+
+            stmt.where(
+                column.in_(
+                    select([othertable.c.y]).
+                    where(table.c.x == othertable.c.x)
+                )
+            )
+
+          In this calling form, :meth:`.ColumnOperators.in_` renders as given::
+
+            WHERE COL IN (SELECT othertable.y
+            FROM othertable WHERE othertable.x = table.x)
+
+        :param other: a list of literals, a :func:`.select` construct,
+         or a :func:`.bindparam` construct that includes the
+         :paramref:`.bindparam.expanding` flag set to True.
 
         """
         return self.operate(in_op, other)
@@ -1167,20 +1249,18 @@ def _escaped_like_impl(fn, other, escape, autoescape):
     if autoescape:
         if autoescape is not True:
             util.warn(
-                "The autoescape parameter is now a simple boolean True/False")
+                "The autoescape parameter is now a simple boolean True/False"
+            )
         if escape is None:
-            escape = '/'
+            escape = "/"
 
         if not isinstance(other, util.compat.string_types):
             raise TypeError("String value expected when autoescape=True")
 
-        if escape not in ('%', '_'):
+        if escape not in ("%", "_"):
             other = other.replace(escape, escape + escape)
 
-        other = (
-            other.replace('%', escape + '%').
-            replace('_', escape + '_')
-        )
+        other = other.replace("%", escape + "%").replace("_", escape + "_")
 
     return fn(other, escape=escape)
 
@@ -1259,13 +1339,24 @@ def json_path_getitem_op(a, b):
 
 _commutative = {eq, ne, add, mul}
 
-_comparison = {eq, ne, lt, gt, ge, le, between_op, like_op, is_,
-               isnot, is_distinct_from, isnot_distinct_from}
+_comparison = {
+    eq,
+    ne,
+    lt,
+    gt,
+    ge,
+    le,
+    between_op,
+    like_op,
+    is_,
+    isnot,
+    is_distinct_from,
+    isnot_distinct_from,
+}
 
 
 def is_comparison(op):
-    return op in _comparison or \
-        isinstance(op, custom_op) and op.is_comparison
+    return op in _comparison or isinstance(op, custom_op) and op.is_comparison
 
 
 def is_commutative(op):
@@ -1273,13 +1364,16 @@ def is_commutative(op):
 
 
 def is_ordering_modifier(op):
-    return op in (asc_op, desc_op,
-                  nullsfirst_op, nullslast_op)
+    return op in (asc_op, desc_op, nullsfirst_op, nullslast_op)
 
 
 def is_natural_self_precedent(op):
-    return op in _natural_self_precedent or \
-        isinstance(op, custom_op) and op.natural_self_precedent
+    return (
+        op in _natural_self_precedent
+        or isinstance(op, custom_op)
+        and op.natural_self_precedent
+    )
+
 
 _booleans = (inv, istrue, isfalse, and_, or_)
 
@@ -1287,12 +1381,8 @@ _booleans = (inv, istrue, isfalse, and_, or_)
 def is_boolean(op):
     return is_comparison(op) or op in _booleans
 
-_mirror = {
-    gt: lt,
-    ge: le,
-    lt: gt,
-    le: ge
-}
+
+_mirror = {gt: lt, ge: le, lt: gt, le: ge}
 
 
 def mirror(op):
@@ -1306,17 +1396,18 @@ def mirror(op):
 
 _associative = _commutative.union([concat_op, and_, or_]).difference([eq, ne])
 
-_natural_self_precedent = _associative.union([
-    getitem, json_getitem_op, json_path_getitem_op])
+_natural_self_precedent = _associative.union(
+    [getitem, json_getitem_op, json_path_getitem_op]
+)
 """Operators where if we have (a op b) op c, we don't want to
 parenthesize (a op b).
 
 """
 
 
-_asbool = util.symbol('_asbool', canonical=-10)
-_smallest = util.symbol('_smallest', canonical=-100)
-_largest = util.symbol('_largest', canonical=100)
+_asbool = util.symbol("_asbool", canonical=-10)
+_smallest = util.symbol("_smallest", canonical=-100)
+_largest = util.symbol("_largest", canonical=100)
 
 _PRECEDENCE = {
     from_: 15,
@@ -1325,7 +1416,6 @@ _PRECEDENCE = {
     getitem: 15,
     json_getitem_op: 15,
     json_path_getitem_op: 15,
-
     mul: 8,
     truediv: 8,
     div: 8,
@@ -1333,22 +1423,17 @@ _PRECEDENCE = {
     neg: 8,
     add: 7,
     sub: 7,
-
     concat_op: 6,
-
     match_op: 5,
     notmatch_op: 5,
-
     ilike_op: 5,
     notilike_op: 5,
     like_op: 5,
     notlike_op: 5,
     in_op: 5,
     notin_op: 5,
-
     is_: 5,
     isnot: 5,
-
     eq: 5,
     ne: 5,
     is_distinct_from: 5,
@@ -1359,7 +1444,6 @@ _PRECEDENCE = {
     lt: 5,
     ge: 5,
     le: 5,
-
     between_op: 5,
     notbetween_op: 5,
     distinct_op: 5,
@@ -1369,17 +1453,14 @@ _PRECEDENCE = {
     and_: 3,
     or_: 2,
     comma_op: -1,
-
     desc_op: 3,
     asc_op: 3,
     collate: 4,
-
     as_: -1,
     exists: 0,
-
     _asbool: -10,
     _smallest: _smallest,
-    _largest: _largest
+    _largest: _largest,
 }
 
 
@@ -1387,7 +1468,6 @@ def is_precedent(operator, against):
     if operator is against and is_natural_self_precedent(operator):
         return False
     else:
-        return (_PRECEDENCE.get(operator,
-                                getattr(operator, 'precedence', _smallest)) <=
-                _PRECEDENCE.get(against,
-                                getattr(against, 'precedence', _largest)))
+        return _PRECEDENCE.get(
+            operator, getattr(operator, "precedence", _smallest)
+        ) <= _PRECEDENCE.get(against, getattr(against, "precedence", _largest))
